/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This is the text-based interface for the PRICE assembler.
 *
 */


#include "AssemblyException.h"
#include "ParamsCycleManagement.h"
#include "Assembler.h"

#include <set>
#include <sstream>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <queue>
#include <string.h>
#include <bitset>
#include <cstddef>
#include <omp.h>
using namespace::std;
using ::size_t;

string helpString = string(
"PRICE v0.11.1\n"
"\n"
"These are the command options for the PRICE assembler. \n"
"For more details, see the accompanying README.txt file. \n"
"Contact: price@derisilab.ucsf.edu \n"
"\n"
"Usage: ./PriceTI [args] \n"
"\n"
"INPUT FILES: \n"
" accepted formats are fasta (.fa or .fasta), fastq (.fq, .fastq, or _sequence.txt), or priceq (.pq or .priceq) \n"
"INPUT READ FILES: \n"
"  NOTE: these flags can be used multiple times in the same command to include multiple read datasets. \n"
"  (default % ID = 90%) \n"
" PAIRED-END FILES (reads are 3p of one another on opposite strands, i.e. pointing towards one another)\n"
" -fp a b c [d e [f]]: (a,b)input file pair, (c)amplicon insert size (including read) \n"
"                  (d,e,f) are optional; (d)the num. cycles to be skipped before this file is used;\n"
"                  if (f) is provided, then the file will alternate between being used for (e) cycles and not used for (f) cycles;\n"
"                  otherwise, the file will be used for (e) cycles then will not be used again. \n"
" -fpp a b c d [e f [g]]: (a,b)input file pair, (c)amplicon insert size (including read), (d)required % identity for match (25-100 allowed) \n"
"                  (e,f,g) are optional; (e)the num. cycles to be skipped before this file is used;\n"
"                  if (g) is provided, then the file will alternate between being used for (f) cycles and not used for (g) cycles;\n"
"                  otherwise, the file will be used for (f) cycles then will not be used again. \n"
" -fs a b [c d [e]]: (a)input paired-end file (alternating entries are paired-end reads), (b)amplicon insert size (including read) \n"
"                  (c,d,e) are optional; (c)the num. cycles to be skipped before this file is used;\n"
"                  if (e) is provided, then the file will alternate between being used for (d) cycles and not used for (e) cycles;\n"
"                  otherwise, the file will be used for (d) cycles then will not be used again. \n"
" -fsp a b c [d e [f]]: (a)input paired-end file (alternating entries are paired-end reads), (b)amplicon insert size (including read),\n"
"                  (c)required % identity for match (25-100 allowed) \n"
"                  (d,e,f) are optional; (d)the num. cycles to be skipped before this file is used;\n"
"                  if (f) is provided, then the file will alternate between being used for (e) cycles and not used for (f) cycles;\n"
"                  otherwise, the file will be used for (e) cycles then will not be used again. \n"
" MATE-PAIR FILES (reads are 5p of one another on opposite strands, i.e. pointing away from one another)\n"
" -mp a b c [d e [f]]: like -fp above, but with reads in the opposite orientation.\n"
" -mpp a b c d [e f [g]]: like -fpp above, but with reads in the opposite orientation.\n"
" -ms a b [c d [e]]: like -fs above, but with reads in the opposite orientation.\n"
" -msp a b c [d e [f]]: like -fsp above, but with reads in the opposite orientation.\n"
"INPUT INITIAL CONTIG FILES: \n"
"  NOTE: these flags can be used multiple times in the same command to include multiple initial contig datasets. \n"
" -icf a b c d: (a)initial contig file, (b)number of addition steps, (c)number of cycles per step, (d)const by which to multiply quality scores \n"
" -picf a b c d e: (a)num of initial contigs from this file, (b)initial contig file, (c)num addition steps, (d)num cycles per step (e)const by which to multiply quality scores \n"
" -icfNt/-picfNt: same as -icf/-picf, but if target mode is invoked, contigs with matches to these input sequences will not necessarily be retained \n"
"OUTPUT FILES: \n"
" accepted formats are fasta (.fa or .fasta) or priceq (.pq or .priceq) \n"
" -o a: (a)output file name (.fasta or .priceq) \n"
" -nco a: (a)num. cycles that pass in between output files being written (default=1) \n"
"OTHER PARAMS: \n"
" -nc a: (a)num. of cycles \n"
" -link a: (a)max. number of contigs that are allowed to replace a read in a mini-assembly (default=5)\n"
" -mol a: (a)minimum overlap length for mini-assembly (default=35) \n"
" -tol a: (a)threshold seq num for scaling overlap for overhang assemblies (default=20)\n"
" NOTE: -mol and -tol do not affect the parameters for de-Bruijn-graph-based assembly.\n"
" -mpi a: (a)minimum % identity for mini-assembly (default=85) \n"
" -tpi a: (a)threshold seq num for scaling % ID for mini-assemblies (default=20)\n"
" -MPI, -TPI : same as above, but for meta-assembly (-MPI default=85, -TPI default=1000)\n"
" NOTE: there is no minimum overlap value for meta-assembly\n"
" -trim a b [c]: after (a) cycles, (b)min. coverage level, (c)optional min. length after trim\n"
"            if (a) == 0, initial contigs will be trimmed whenever they are added\n"
" -target a b [c d]: limit output contigs to those with matches to input initial contigs at the end of each cycle.\n"
"                    (a) % identity to an input initial contig to count as a match (ungapped); (b)num cycles to skip\n"
"                    before applying this filter.  [c and d are optional, but must both be provided if either is]\n"
"                    After target filtering has begin, target-filtered/-unfiltered cycles will alternate with (c)\n"
"                    filtered cycles followed by (d) unfiltered cycles.\n"
" -targetF a b [c d]: the same as -target, but now matches to all reads in the input set will be specified, not just\n"
"                    the ones that have been introduced up to that point (this is FullFile mode).\n"
" -dbmax a: (a) the maximum length sequence that will be fed into de Bruijn assembly \n"
"           (default=100; recommended: max paired-end read length)\n"
" -dbk a: (a) the k-mer size for de Bruijn assembly (default=20; keep less than the read length)\n"
" -dbms a: (a) the minimum number of sequences to which de Bruijn assembly will be applied (default=3)\n"
" -lenf a b: filters out contigs shorter than (a) nt at the end of every cycle, after skipping (b) cycles.\n"
" -maxHp a: filters out a pair of reads if either read has a homo-polymer track >(a) nucleotides in length.\n"
" -badf a b: prevents reads with an ungapped match of at least (b)% identity to a sequence in file (a) from being mapped to contigs.\n"
" -reset a [b c d...]: re-introduces contigs that were previously not generating assembly jobs of their own\n"
"                      (a) is the one-indexed cycle where the contigs will be reset.  Same with b, c, d. \n"
"                      Any number of args may be added.\n"
"COMPUTATIONAL EFFICIENCY: \n"
" -a x: (x)num threads to use (default=1) \n"
" -mtpf a: (a)max threads per file (default=1) \n"
"USER INTERFACE: \n"
" -log a: determines the type of outputmakes the output verbose (lots of time stamp tags) \n"
"         (a) = c: concise stdout (default)\n"
"         (a) = n: no stdout \n"
"         (a) = v: verbose stdout \n"
" -logf a: (a)the name of an output file for verbose log info to be written (doesn't change stdout format) \n"
" -, -h, or --help: user interface info. \n"
);

int main(int argc, char *argv[]){

  int initialInput = 0;
  bool willDoAssembly = true; // allows full set of errors to be printed
  bool distributeInitialContigs = false;
  int cyclesPerStep;
  int numSteps;

  // default is 1 core
  omp_set_num_threads( 1 );
  int maxThreadsPerFile = 1;

  // these may be changed
  ParamsMinOverlap* pmoDefault = Assembler::getDefaultParamsMinOverlap();
  ParamsMinFractId* pmfDefault = ParamsMinFractId::getDefault();

  ParamsMinOverlap* pmoMini = Assembler::getDefaultParamsMinOverlap();
  ParamsMinFractId* pmfMini = ParamsMinFractId::getDefault();
  AlignmentScoreMatrix* asmMini = AlignmentScoreMatrix::getDefault();

  ParamsMinOverlap* pmoMeta = new ParamsMinOverlap(200,
						   pmoDefault->minOvlThreshold(),
						   pmoDefault->minOvlSlope());
  
  ParamsMinFractId* pmfMeta = new ParamsMinFractId(pmfDefault->globalFractId(),
						   1000,
						   pmfDefault->denominatorLogStep());
  AlignmentScoreMatrix* asmMeta = AlignmentScoreMatrix::getDefault();
  delete pmoDefault;
  delete pmfDefault;

  bool listenerIsNull = false;
  bool listenerIsVerbose = false;
  bool includeListenerFile = false;
  string listenerFilename = "";

  // first, get the total number of cycles and create the cycle manager
  int numOfCyclesToDo = 0;
  long maxNumLinks = ParamsCycleManagement::defaultMaxContigLinkages();
  bool hasOutputInterval = false;
  int outputInterval;
  int argn = 1;
  while ( argn < argc ){
    string arg = string(argv[argn]);
    if (arg == "-nc"){
      numOfCyclesToDo = atoi( argv[argn+1] );
      argn += 2;
    } else if (arg == "-link"){
      maxNumLinks = atol( argv[argn+1] );
      argn += 2;
    } else if (arg == "-nco"){
      hasOutputInterval = true;
      outputInterval = atoi( argv[argn+1] );
      argn += 2;
    } else { argn++; }
  }
  ParamsCycleManagement* cycleInfo = new ParamsCycleManagement(numOfCyclesToDo,maxNumLinks);
  if (hasOutputInterval){ cycleInfo->setOutputInterval(outputInterval); }
  argn = 1;
  while (argn < argc){
    string arg = string(argv[argn]);
    if (arg == "-reset"){
      int numResets = 0;
      argn += 1;
      while (argn < argc and argv[argn][0] != '-'){
	cycleInfo->addContigReset( atoi( argv[argn] ) );
	argn += 1;
	numResets++;
      }
      if (numResets == 0 ){
	throw AssemblyException::ArgError("-reset should have at least one cycle number after it.");
      }
    } else { argn += 1; }
  }
  Assembler* assembler = new Assembler(cycleInfo);



  // now go through again and get the command-line args organized
  argn = 1;
  while ( argn < argc ){
    string arg = string(argv[argn]);
    if (arg == "-fp"){
      string fileA = argv[argn+1];
      string fileB = argv[argn+2];
      long ampSize = atol(argv[argn+3]);
      if ( argn + 4 >= argc or argv[argn + 4][0] == '-'){
	assembler->addReadFile(fileA,fileB,ampSize);
	argn += 4;
      } else if ( argn + 6 >= argc or argv[argn + 6][0] == '-'){
	int startCycle = atoi(argv[argn+4]);
	int numCycles = atoi(argv[argn+5]);
	assembler->addReadFileLimitUse(startCycle,numCycles,fileA,fileB,ampSize);
	argn += 6;
      } else {
	int startCycle = atoi(argv[argn+4]);
	int onCycles = atoi(argv[argn+5]);
	int skipCycles = atoi(argv[argn+6]);
	assembler->addReadFileLimitUse(startCycle,onCycles,skipCycles,fileA,fileB,ampSize);
	argn += 7;
      }
    } else if (arg == "-fpp"){
      string fileA = argv[argn+1];
      string fileB = argv[argn+2];
      long ampSize = atol(argv[argn+3]);

      // helps deal with the optional last two args
      float fractId = float(atoi(argv[argn+4])) / float(100);
      if ( argn + 5 >= argc or argv[argn + 5][0] == '-'){
	assembler->addReadFile(fileA,fileB,ampSize,fractId);
	argn += 5;
      } else if ( argn + 7 >= argc or argv[argn + 7][0] == '-'){
	int startCycle = atoi(argv[argn+5]);
	int numCycles = atoi(argv[argn+6]);
	assembler->addReadFileLimitUse(startCycle,numCycles,fileA,fileB,ampSize);
	argn += 7;
      } else {
	int startCycle = atoi(argv[argn+5]);
	int onCycles = atoi(argv[argn+6]);
	int skipCycles = atoi(argv[argn+7]);
	assembler->addReadFileLimitUse(startCycle,onCycles,skipCycles,fileA,fileB,ampSize);
	argn += 8;
      }
    } else if (arg == "-fs"){
      string file = argv[argn+1];
      long ampSize = atol(argv[argn+2]);
      if ( argn + 3 >= argc or argv[argn + 3][0] == '-'){
	assembler->addReadFile(file,ampSize);
	argn += 3;
      } else if ( argn + 5 >= argc or argv[argn + 5][0] == '-'){
	int startCycle = atoi(argv[argn+3]);
	int numCycles = atoi(argv[argn+4]);
	assembler->addReadFileLimitUse(startCycle,numCycles,file,ampSize);
	argn += 5;
      } else {
	int startCycle = atoi(argv[argn+3]);
	int onCycles = atoi(argv[argn+4]);
	int skipCycles = atoi(argv[argn+5]);
	assembler->addReadFileLimitUse(startCycle,onCycles,skipCycles,file,ampSize);
	argn += 6;
      }
    } else if (arg == "-fsp"){
      string file = argv[argn+1];
      long ampSize = atol(argv[argn+2]);

      // helps deal with the optional last two args
      float fractId = float(atoi(argv[argn+3])) / float(100);
      if ( argn + 4 >= argc or argv[argn + 4][0] == '-'){
	assembler->addReadFile(file,ampSize,fractId);
	argn += 4;
      } else if ( argn + 6 >= argc or argv[argn + 6][0] == '-'){
	int startCycle = atoi(argv[argn+4]);
	int numCycles = atoi(argv[argn+5]);
	assembler->addReadFileLimitUse(startCycle,numCycles,file,ampSize);
	argn += 6;
      } else {
	int startCycle = atoi(argv[argn+4]);
	int onCycles = atoi(argv[argn+5]);
	int skipCycles = atoi(argv[argn+6]);
	assembler->addReadFileLimitUse(startCycle,onCycles,skipCycles,file,ampSize);
	argn += 7;
      }
    } else if (arg == "-mp"){
      string fileA = argv[argn+1];
      string fileB = argv[argn+2];
      long ampSize = atol(argv[argn+3]);
      if ( argn + 4 >= argc or argv[argn + 4][0] == '-'){
	assembler->addMatePairFile(fileA,fileB,ampSize);
	argn += 4;
      } else if ( argn + 6 >= argc or argv[argn + 6][0] == '-'){
	int startCycle = atoi(argv[argn+4]);
	int numCycles = atoi(argv[argn+5]);
	assembler->addMatePairFileLimitUse(startCycle,numCycles,fileA,fileB,ampSize);
	argn += 6;
      } else {
	int startCycle = atoi(argv[argn+4]);
	int onCycles = atoi(argv[argn+5]);
	int skipCycles = atoi(argv[argn+6]);
	assembler->addMatePairFileLimitUse(startCycle,onCycles,skipCycles,fileA,fileB,ampSize);
	argn += 7;
      }
    } else if (arg == "-mpp"){
      string fileA = argv[argn+1];
      string fileB = argv[argn+2];
      long ampSize = atol(argv[argn+3]);

      // helps deal with the optional last two args
      float fractId = float(atoi(argv[argn+4])) / float(100);
      if ( argn + 5 >= argc or argv[argn + 5][0] == '-'){
	assembler->addMatePairFile(fileA,fileB,ampSize,fractId);
	argn += 5;
      } else if ( argn + 7 >= argc or argv[argn + 7][0] == '-'){
	int startCycle = atoi(argv[argn+5]);
	int numCycles = atoi(argv[argn+6]);
	assembler->addMatePairFileLimitUse(startCycle,numCycles,fileA,fileB,ampSize);
	argn += 7;
      } else {
	int startCycle = atoi(argv[argn+5]);
	int onCycles = atoi(argv[argn+6]);
	int skipCycles = atoi(argv[argn+7]);
	assembler->addMatePairFileLimitUse(startCycle,onCycles,skipCycles,fileA,fileB,ampSize);
	argn += 8;
      }
    } else if (arg == "-ms"){
      string file = argv[argn+1];
      long ampSize = atol(argv[argn+2]);
      if ( argn + 3 >= argc or argv[argn + 3][0] == '-'){
	assembler->addMatePairFile(file,ampSize);
	argn += 3;
      } else if ( argn + 5 >= argc or argv[argn + 5][0] == '-'){
	int startCycle = atoi(argv[argn+3]);
	int numCycles = atoi(argv[argn+4]);
	assembler->addMatePairFileLimitUse(startCycle,numCycles,file,ampSize);
	argn += 5;
      } else {
	int startCycle = atoi(argv[argn+3]);
	int onCycles = atoi(argv[argn+4]);
	int skipCycles = atoi(argv[argn+5]);
	assembler->addMatePairFileLimitUse(startCycle,onCycles,skipCycles,file,ampSize);
	argn += 6;
      }
    } else if (arg == "-msp"){
      string file = argv[argn+1];
      long ampSize = atol(argv[argn+2]);

      // helps deal with the optional last two args
      float fractId = float(atoi(argv[argn+3])) / float(100);
      if ( argn + 4 >= argc or argv[argn + 4][0] == '-'){
	assembler->addMatePairFile(file,ampSize,fractId);
	argn += 4;
      } else if ( argn + 6 >= argc or argv[argn + 6][0] == '-'){
	int startCycle = atoi(argv[argn+4]);
	int numCycles = atoi(argv[argn+5]);
	assembler->addMatePairFileLimitUse(startCycle,numCycles,file,ampSize);
	argn += 6;
      } else {
	int startCycle = atoi(argv[argn+4]);
	int onCycles = atoi(argv[argn+5]);
	int skipCycles = atoi(argv[argn+6]);
	assembler->addMatePairFileLimitUse(startCycle,onCycles,skipCycles,file,ampSize);
	argn += 7;
      }
    } else if (arg == "-icf"){
      string filename = argv[argn+1];
      float countFactor = atof( argv[argn+4] );
      int numSteps = atoi( argv[argn+2] );
      int cyclesPerStep = atoi( argv[argn+3] );
      assembler->addInitialFile(filename,numSteps,cyclesPerStep,countFactor);
      argn += 5;
    } else if (arg == "-picf"){
      long initialInput = atol( argv[argn+1] );
      string filename = argv[argn+2];
      float countFactor = atof( argv[argn+5] );
      int numSteps = atoi( argv[argn+3] );
      int cyclesPerStep = atoi( argv[argn+4] );
      assembler->addInitialFile(initialInput,filename,numSteps,cyclesPerStep,countFactor);
      argn += 6;
   } else if (arg == "-icfNt"){
      string filename = argv[argn+1];
      float countFactor = atof( argv[argn+4] );
      int numSteps = atoi( argv[argn+2] );
      int cyclesPerStep = atoi( argv[argn+3] );
      assembler->addInitialFileNoTarget(filename,numSteps,cyclesPerStep,countFactor);
      argn += 5;
    } else if (arg == "-picfNt"){
      long initialInput = atol( argv[argn+1] );
      string filename = argv[argn+2];
      float countFactor = atof( argv[argn+5] );
      int numSteps = atoi( argv[argn+3] );
      int cyclesPerStep = atoi( argv[argn+4] );
      assembler->addInitialFileNoTarget(initialInput,filename,numSteps,cyclesPerStep,countFactor);
      argn += 6;
    } else if (arg == "-nc"){
      // nothing happens here because this info was already collected
      argn += 2;
    } else if (arg == "-target" or arg == "-targetF"){
      bool fullFileMode = (arg == "-targetF");
      float fractId = atof( argv[argn+1] ) / float(100.0);
      int cyclesToSkip = atoi( argv[argn+2] );
      if ( argn + 3 >= argc or argv[argn+3][0] == '-'){
	assembler->addTargetMode(fractId,cyclesToSkip,fullFileMode);
	argn += 3;
      } else {
	int numFilterCycles = atoi( argv[argn+3] );
	int numSkipCycles = atoi( argv[argn+4] );
	assembler->addTargetMode(fractId,cyclesToSkip,numFilterCycles,numSkipCycles,fullFileMode);
	argn += 5;
      }
    } else if (arg == "-maxHp"){
      assembler->addHomopolymerFilter( atol(argv[argn+1]) );
      argn += 2;
    } else if (arg == "-badf"){
      string filename = argv[argn+1];
      //FileType filetype = getFileType(filename);
      float fractId = atof( argv[argn+2] ) / float(100.0);
      assembler->addBadSequenceFilter(filename, fractId);
      argn += 3;
    } else if (arg == "-lenf"){
      long minLength = atol( argv[argn+1] );
      int cyclesToSkip = atoi( argv[argn+2] );
      assembler->addLengthFilter(minLength, cyclesToSkip);
      argn += 3;
    } else if (arg == "-link"){
      // nothing happens; already dealt with
      argn += 2;
    } else if (arg == "-o"){
      string filename = argv[argn+1];
      assembler->addOutputFile(filename);
      argn += 2; 
    } else if (arg == "-nco"){
      // nothing happens here because this info was already collected
      argn += 2;
    } else if (arg == "-a"){
      omp_set_num_threads( atoi( argv[argn+1] ) );
      argn += 2; 
    } else if (arg == "-dbk"){
      assembler->setDeBruijnKmerSize( atoi( argv[argn+1] ) );
      argn += 2; 
    } else if (arg == "-dbmax"){
      assembler->setDeBruijnMaxSeqLength( atol( argv[argn+1] ) );
      argn += 2; 
    } else if (arg == "-dbms"){
      assembler->setDeBruijnMinSeqNum( atol( argv[argn+1] ) );
      argn += 2; 
    } else if (arg == "-mtpf"){
      maxThreadsPerFile = atoi( argv[argn+1] );
      argn += 2; 
    } else if (arg == "-trim"){
      int numCycles = atoi( argv[argn+1] );
      float minCoverage = atof( argv[argn+2] );
      if ( argn + 3 < argc and argv[argn+3][0] != '-' ) {
	long minLength = atol( argv[argn+3] );
	cycleInfo->addTrim(numCycles, minCoverage, minLength);
	argn += 4;
      } else {
	cycleInfo->addTrim(numCycles, minCoverage);
	argn += 3;
      }
    } else if (arg == "-reset"){
      int numResets = 0;
      argn += 1;
      while (argn < argc and argv[argn][0] != '-'){
	// this was already taken care of so i don't need to do anything here
	//cycleInfo->addContigReset( atoi( argv[argn] ) );
	argn += 1;
	numResets++;
      }
      if (numResets == 0 ){
	throw AssemblyException::ArgError("-reset should have at least one cycle number after it.");
      }
    } else if (arg == "-mol"){ // min overlap for overhang assemblies
      long minOvl = atol( argv[argn+1] );
      ParamsMinOverlap* pmoMiniHolder = new ParamsMinOverlap(minOvl,
							     pmoMini->minOvlThreshold(),
							     pmoMini->minOvlSlope());
      delete pmoMini;
      pmoMini = pmoMiniHolder;
      argn += 2; 
    } else if (arg == "-tol"){ // threshold for scaling overlap for overhang assemblies
      long ovlThreshold = atol( argv[argn+1] );
      ParamsMinOverlap* pmoMiniHolder = new ParamsMinOverlap(pmoMini->globalMinOverlap(),
							     ovlThreshold,
							     pmoMini->minOvlSlope());
      delete pmoMini;
      pmoMini = pmoMiniHolder;
      argn += 2; 
    } else if (arg == "-mpi"){ // min percent ID for overhang assemblies
      int pctId = atoi(argv[argn+1]);
      if (pctId < 25 or pctId > 100){
	throw AssemblyException::ArgError("% ID required for mini-assembly is out of range 25-100");
      }
      ParamsMinFractId* pmfMiniHolder = new ParamsMinFractId(float(pctId) / 100.0,
							     pmfMini->fractIdThreshold(),
							     pmfMini->denominatorLogStep());
      delete pmfMini;
      pmfMini = pmfMiniHolder;
      argn += 2; 
    } else if (arg == "-tpi"){ // threshold for scaling percent ID for overhang assemblies
      long threshold = atol(argv[argn+1]);
      ParamsMinFractId* pmfMiniHolder = new ParamsMinFractId(pmfMini->globalFractId(),
							     threshold,
							     pmfMini->denominatorLogStep());
      delete pmfMini;
      pmfMini = pmfMiniHolder;
      argn += 2; 
    } else if (arg == "-MPI"){ // min percent ID for meta assemblies
      int pctId = atoi(argv[argn+1]);
      if (pctId < 25 or pctId > 100){
	throw AssemblyException::ArgError("% ID required for mini-assembly is out of range 25-100");
      }
      ParamsMinFractId* pmfMetaHolder = new ParamsMinFractId(float(pctId) / 100.0,
							     pmfMeta->fractIdThreshold(),
							     pmfMeta->denominatorLogStep());
      delete pmfMeta;
      pmfMeta = pmfMetaHolder;
      argn += 2; 
    } else if (arg == "-TPI"){ // threshold for scaling percent ID for meta assemblies
      long threshold = atol(argv[argn+1]);
      ParamsMinFractId* pmfMetaHolder = new ParamsMinFractId(pmfMeta->globalFractId(),
							     threshold,
							     pmfMeta->denominatorLogStep());
      delete pmfMeta;
      pmfMeta = pmfMetaHolder;
      argn += 2; 
   } else if (arg == "-log"){
      string type = string(argv[argn+1]);
      if (type == "n"){ listenerIsNull = true; }
      else if (type == "c"){ }
      else if (type == "v"){ listenerIsVerbose = true; }
      else { cout << "'" << type << "' is not a valid stdout type ('n', 'c', or 'v')." << endl; }
      argn += 2;
   } else if (arg == "-logf"){
      includeListenerFile = true;
      listenerFilename = argv[argn+1];
      argn += 2;
   } else if (arg == "-h" or arg == "-" or arg == "--help"){
      willDoAssembly = false;
      cout << helpString << endl;
      argn++;
    } else {
      cout << "'" << arg << "' is an invalid argument." << endl;
      throw AssemblyException::ArgError("died due to bad arg flag (see stdout above).");
      argn++;
    }
    if (argn != argc and argv[argn][0] != '-'){
      ostringstream outMessage;
      outMessage << arg << " was given an incorrect number of args.";
      throw AssemblyException::ArgError( outMessage.str().c_str() );
    }
  }

  // TEMPORARY set-up for returning the command that launched the job
  ostringstream launchCommand;
  launchCommand << "Launch command:";
  for (int n = 0; n < argc; ++n){ launchCommand << " " << argv[n]; }
  cout << launchCommand.str().c_str() << endl;

  if ( willDoAssembly ){

    // do checks to make sure all of the initial values are legit???

    ParamsAlignment* miniAlignParams = new ParamsAlignment(pmoMini, pmfMini, asmMini);
    // NOTE: pmoMeta is just a filler; this class should be able to take a NULL for
    // that variable if no min overlap will be used
    ParamsAlignment* metaAlignParams = new ParamsAlignment(pmoMeta, pmfMeta, asmMeta);
    assembler->setMiniAlignParams(miniAlignParams);
    assembler->setMetaAlignParams(metaAlignParams);
    delete miniAlignParams;
    delete metaAlignParams;

    if (listenerIsNull){ assembler->makeLogNull(); }
    else if (listenerIsVerbose){ assembler->makeLogVerbose(); }
    if (includeListenerFile = true){ assembler->verboseLogFile(listenerFilename); }

    if (maxThreadsPerFile != 1){ assembler->setMaxThreadsPerFile(maxThreadsPerFile); }

    assembler->runAssembly();
    delete assembler;

  }
  delete cycleInfo;
  delete pmoMini;
  delete pmfMini;
  delete asmMini;
  delete pmoMeta;
  delete pmfMeta;
  delete asmMeta;
  return 0;
}



